#include <msp.h>
#include "uart.h"
#include "fifo_queue.h"
#include "fifo_function_queue.h"
#include "common.h"

//Reception and Transmission Queues
static char rx_buffer[QUEUE_DIMENSION_UART];
static char tx_buffer[QUEUE_DIMENSION_UART];
static char tail_rx, head_rx;
static char tail_tx, head_tx;

//Callback function for queue
void (*message_received_callback_uart)();

void initializeUART()
{
    //Keep module in reset while being configured
    EUSCI_A0 -> CTLW0 = EUSCI_A_CTLW0_SWRST;

    //Parity disabled, LSB first, 8-bit length, 1 stop bit, asynchronous uart mode
    //SMCLK
    EUSCI_A0 -> CTLW0 |= EUSCI_A_CTLW0_SSEL__SMCLK;

    //12MHz to 115200bps
    EUSCI_A0 -> MCTLW = 0x2000 | 0x0080 | EUSCI_A_MCTLW_OS16;
    EUSCI_A0 -> BRW = 6;

    //Configure shared i/o pins
    //P1.2 Rx as input pin (P1SEL1 = 0, P1SEL0 = 1)
    //P1.3 Tx as output pin (P1SEL1 = 0, P1SEL0 = 1)
    P1SEL0 |= BIT2 | BIT3;
    P1SEL1 &= (~BIT2) & (~BIT3);

    //Init queues
    initQueue(&head_rx, &tail_rx);
    initQueue(&head_tx, &tail_tx);

    //Turn on the module
    EUSCI_A0 -> CTLW0 &= ~EUSCI_A_CTLW0_SWRST;

    //Enable RX Interrupts

    EUSCI_A0 -> IE |= EUSCI_A_IE_RXIE;
    NVIC_SetPriority(EUSCIA0_IRQn,3);
    NVIC_EnableIRQ(EUSCIA0_IRQn);
}

void setUartCallbackFunction(void *callback_function)
{
    message_received_callback_uart = callback_function;
}

void uartReceive(char* buffer_pointer, int* bytes_received)
{
    *bytes_received = 0;
    char value;

    if(!queueIsEmpty(head_rx, tail_rx))
    {
        value = getFromQueue(&tail_rx, rx_buffer);

        while (value != UART_TRAILING_CHARACTER)
        {
            *(buffer_pointer + *bytes_received) = value;

            if(!queueIsEmpty(head_rx, tail_rx))
            {
                value = getFromQueue(&tail_rx, rx_buffer);
            }
            else
            {
                value = 0;
            }

            (*bytes_received)++;
        }
    }
}

void uartTransmit(char* buffer_pointer)
{
    int i = 0;

    while (*(buffer_pointer + i)  != END_CHARACTER)
    {
        if(!queueIsFull(head_tx, tail_tx))
        {
            addToQueue(*(buffer_pointer + i), &head_tx, tx_buffer);
        }
        i++;
    }

    if(!queueIsFull(head_tx, tail_tx))
    {
        addToQueue(*(buffer_pointer + i), &head_tx, tx_buffer);     //Adding END_CHARACTER
    }

    if(!queueIsFull(head_tx, tail_tx))
    {
        addToQueue(UART_TRAILING_CHARACTER, &head_tx, tx_buffer);     //Adding trailing character
    }

    EUSCI_A0 -> IE |= EUSCI_A_IE_TXIE;
}

void EUSCIA0_IRQHandler(void)
{
    int received_character;
    if(EUSCI_A0 -> IFG & EUSCI_A_IFG_RXIFG)
    {
        //ISR - Rx
        received_character = EUSCI_A0 -> RXBUF;
        if(!queueIsFull(head_rx, tail_rx))
        {
            addToQueue(received_character, &head_rx, rx_buffer);

            if(received_character  == UART_TRAILING_CHARACTER)
            {
                addToFunctionQueue(message_received_callback_uart);
                __low_power_mode_off_on_exit();
            }
        }
    }
    else if(EUSCI_A0 -> IFG & EUSCI_A_IFG_TXIFG)
    {
        //ISR - Tx
        if(!queueIsEmpty(head_tx, tail_tx))
        {
            EUSCI_A0 -> TXBUF = getFromQueue(&tail_tx, tx_buffer);
        }
        else
        {
            EUSCI_A0 -> IE &= ~EUSCI_A_IE_TXIE;
        }
    }
}
