#include <msp.h>
#include "uartBluetooth.h"
#include "fifo_queue.h"
#include "fifo_function_queue.h"
#include "common.h"

//Reception and Transmission Queues
static char rx_buffer[QUEUE_DIMENSION_UART];
static char tx_buffer[QUEUE_DIMENSION_UART];
static char tail_rx, head_rx;
static char tail_tx, head_tx;

//Callback function for queue
void (*message_received_callback_bt)();

void initializeUARTBluetooth()
{
    //Keep module in reset while being configured
    EUSCI_A2 -> CTLW0 = EUSCI_A_CTLW0_SWRST;

    //Parity disabled, LSB first, 8-bit length, 1 stop bit, asynchronous uart mode
    //SMCLK
    EUSCI_A2 -> CTLW0 |= EUSCI_A_CTLW0_SSEL__SMCLK;

    if(BLUETOOTH_CONFIGURED == 1)
    {
        //TODO Increase it to 460800bps (could be higher but the SMCLK would need to be increased as well)
        //12MHz to 115200bps
        //EUSCI_A2 -> MCTLW = 0x2000 | 0x0080 | EUSCI_A_MCTLW_OS16;
        //EUSCI_A2 -> BRW = 6;

        //12MHz to 9600bps
        EUSCI_A2 -> MCTLW = 0x0020 | EUSCI_A_MCTLW_OS16;
        EUSCI_A2 -> BRW = 78;
    }
    else
    {
        //12MHz to 9600bps
        EUSCI_A2 -> MCTLW = 0x0020 | EUSCI_A_MCTLW_OS16;
        EUSCI_A2 -> BRW = 78;
    }

    //Configure shared i/o pins
    //P3.2 Rx as input pin (P1SEL1 = 0, P1SEL0 = 1)
    //P3.3 Tx as output pin (P1SEL1 = 0, P1SEL0 = 1)
    P3SEL0 |= BIT2 | BIT3;
    P3SEL1 &= (~BIT2) & (~BIT3);

    //Init queues
    initQueue(&head_rx, &tail_rx);
    initQueue(&head_tx, &tail_tx);

    //Turn on the module
    EUSCI_A2 -> CTLW0 &= ~EUSCI_A_CTLW0_SWRST;

    //Enable RX Interrupts

    EUSCI_A2 -> IE |= EUSCI_A_IE_RXIE;
    NVIC_SetPriority(EUSCIA2_IRQn,3);
    NVIC_EnableIRQ(EUSCIA2_IRQn);
}

void setUartBluetoothCallbackFunction(void *callback_function)
{
    message_received_callback_bt = callback_function;
}

void uartBluetoothReceive(char* buffer_pointer, int* bytes_received)
{
    *bytes_received = 0;
    char value;

    if(!queueIsEmpty(head_rx, tail_rx))
    {
        value = getFromQueue(&tail_rx, rx_buffer);

        while (value != BT_TRAILING_CHARACTER_1)
        {
            *(buffer_pointer + *bytes_received) = value;

            if(!queueIsEmpty(head_rx, tail_rx))
            {
                value = getFromQueue(&tail_rx, rx_buffer);
            }
            else
            {
                value = 0;
            }

            (*bytes_received)++;
        }
    }
}

void uartBluetoothTransmit(char* buffer_pointer)
{
    int i = 0;

    while (*(buffer_pointer + i)  != END_CHARACTER)
    {
        if(!queueIsFull(head_tx, tail_tx))
        {
            addToQueue(*(buffer_pointer + i), &head_tx, tx_buffer);
        }
        i++;
    }

    if(!queueIsFull(head_tx, tail_tx))
    {
        addToQueue(*(buffer_pointer + i), &head_tx, tx_buffer);     //Adding communication ending character
    }
    if(!queueIsFull(head_tx, tail_tx))
    {
        addToQueue(BT_TRAILING_CHARACTER_1, &head_tx, tx_buffer);     //Adding trailing character
    }
    if(!queueIsFull(head_tx, tail_tx))
    {
        addToQueue(BT_TRAILING_CHARACTER_2, &head_tx, tx_buffer);     //Adding trailing character
    }

    EUSCI_A2 -> IE |= EUSCI_A_IE_TXIE;
}

void EUSCIA2_IRQHandler(void)
{
    int received_character;
    if(EUSCI_A2 -> IFG & EUSCI_A_IFG_RXIFG)
    {
        //ISR - Rx
        received_character = EUSCI_A2 -> RXBUF;
        if(!queueIsFull(head_rx, tail_rx))
        {
            if(received_character == BT_TRAILING_CHARACTER_2)
            {
                addToFunctionQueue(message_received_callback_bt);
                __low_power_mode_off_on_exit();
            }
            else
            {
                //Do not put BT_TRAILING_CHARACTER_2 into the queue
                addToQueue(received_character, &head_rx, rx_buffer);
            }
        }
    }
    else if(EUSCI_A2 -> IFG & EUSCI_A_IFG_TXIFG)
    {
        //ISR - Tx
        if(!queueIsEmpty(head_tx, tail_tx))
        {
            EUSCI_A2 -> TXBUF = getFromQueue(&tail_tx, tx_buffer);
        }
        else
        {
            EUSCI_A2 -> IE &= ~EUSCI_A_IE_TXIE;
        }
    }
}
